#!/usr/bin/env python
"""
vault secret store rest api client.

Vault is used to protect secrets in a secure way
using a golang restful web api.  This client is
a wrapper to enable authentication and restful
interactions with the vault rest endpoint.
https://github.com/hashicorp/vault
"""
import os
import sys

import requests
import logging

from vault_client import exceptions

__author__ = "Gary Wright<zerosignal0@msn.com>"
__contributors__ = ["Gary Wright"]
__version__ = "1.0.0"
__maintainer__ = "Gary Wright"
__team_email__ = "zerosignal0@msn.com"
__script_name__ = os.path.basename(sys.argv[0])

API_VERSION = 'v1'

def raise_for_status(func):
    def decorator(*args, **kwargs):
        response = func(*args, **kwargs)

        if response.status_code >= 400 and response.status_code < 600:
            errors = response.json().get('errors')

            if response.status_code == 400:
                raise exceptions.InvalidRequest(errors=errors)
            elif response.status_code == 401:
                raise exceptions.Unauthorized(errors=errors)
            elif response.status_code == 404:
                raise exceptions.InvalidPath(errors=errors)
            elif response.status_code == 429:
                raise exceptions.RateLimitExceeded(errors=errors)
            elif response.status_code == 500:
                raise exceptions.InternalServerError(errors=errors)
            elif response.status_code == 503:
                raise exceptions.VaultDown(errors=errors)
            else:
                raise exceptions.UnknownError()

        return response

    return decorator

class Client(object):
    """
    Client handles the interactions
    required to auth with the vault rest
    api, along with providing access to sys/secrets endpoints.
    You can see an overview of the process here:
        https://www.vaultproject.io/docs/

    TODO:
        Add more authentication types to the client. Currently
        only token / app-id are supported.
    """

    def __init__(self, url, token=''):
        self._url = url
        self._token = token

    @raise_for_status
    def _get(self, url, **kwargs):
        return requests.get(self._url + url, cookies=self._cookies, **kwargs)

    @raise_for_status
    def _post(self, url, data=None, **kwargs):
        return requests.post(self._url + url, json=data, cookies=self._cookies, **kwargs)

    @raise_for_status
    def _put(self, url, data=None, **kwargs):
        return requests.put(self._url + url, json=data, cookies=self._cookies, **kwargs)

    @raise_for_status
    def _delete(self, url, **kwargs):
        return requests.delete(self._url + url, cookies=self._cookies, **kwargs)

    @property
    def _cookies(self):
        return {'token': self._token}

    #######
    # MAIN HTTP VERB ACTIONS
    def read(self, path):
        """
        GET /<path>
        """
        return self._get('/{}/{}'.format(API_VERSION,
                                         path)).json()

    def write(self, path, **kwargs):
        """
        PUT /<path>
        """
        self._put('/{}/{}'.format(API_VERSION,
                                  path), kwargs)

    def delete(self, path):
        """
        DELETE /<path>
        """
        self._delete('/{}/{}'.format(API_VERSION,
                                     path))

    #######
    # TOKEN INTERACTIONS
    def create_token(self, id=None, policies=None, metadata=None,
                     no_parent=False, lease=None, display_name=None,
                     num_uses=None):
        """
        POST /auth/token/create

        https://vaultproject.io/docs/auth/token.html
        """
        params = {
            'id': id,
            'policies': policies,
            'metadata': metadata,
            'no_parent': no_parent,
            'lease': lease,
            'display_name': display_name,
            'num_uses': num_uses,
        }

        return self._post('/{}/auth/token/create'.format(
            API_VERSION), params).json()

    def lookup_token(self, token=None):
        """
        GET /auth/token/lookup/<token>
        GET /auth/token/lookup-self

        https://vaultproject.io/docs/auth/token.html
        """
        if token:
            return self._get('/{}/auth/token/lookup/{}'.format(API_VERSION,
                                                               token)).json()
        else:
            return self._get('/{}/auth/token/lookup-self'.format(
                API_VERSION)).json()

    def revoke_token(self, token, orphan=False):
        """
        POST /auth/token/revoke/<token>
        POST /auth/token/revoke-orphan/<token>

        https://vaultproject.io/docs/auth/token.html
        """
        if orphan:
            self._post('/{}/auth/token/revoke-orphan/{}'.format(API_VERSION,
                                                                token))
        else:
            self._post('/{}/auth/token/revoke/{}'.format(API_VERSION,
                                                         token))

    def revoke_token_prefix(self, prefix):
        """
        POST /auth/token/revoke-prefix/<prefix>

        https://vaultproject.io/docs/auth/token.html
        """
        self._post('/{}/auth/token/revoke-prefix/{}'.format(API_VERSION,
                                                            prefix))

    def renew_token(self, token, increment=None):
        """
        POST /auth/token/renew/<token>

        https://vaultproject.io/docs/auth/token.html
        """
        params = {
            'increment': increment,
        }

        return self._post('/{}/auth/token/renew/{}'.format(API_VERSION,
            token), params).json()

    #######
    # APP-ID INTERACTIONS
    def auth_app_id(self, app_id, user_id, change_to=True, mount_point='app-id'):
        """
        POST /auth/<mount point>/login

        https://vaultproject.io/docs/auth/app-id.html
        """
        params = {
            'app_id': app_id,
            'user_id': user_id,
        }

        result = self._post('/{}/auth/{}/login'.format(API_VERSION,
            mount_point), params).json()

        if change_to:
            self._token = result['auth']['client_token']

        return result

    def auth_userpass(self, username, password, change_to=True, mount_point='userpass'):
        """
        POST /auth/<mount point>/login/<username>

        https://vaultproject.io/docs/auth/userpass.html
        """
        params = {
            'password': password,
        }

        result = self._post('/{}/auth/{}/login/{}'.format(API_VERSION,
            mount_point, username), params).json()

        if change_to:
            self._token = result['auth']['client_token']

        return result

    #######
    # AUTH BACKEND INTERACTIONS
    def list_auth_backends(self):
        """
        GET /sys/auth

        https://vaultproject.io/docs/http/sys-auth.html
        """
        return self._get('/{}/sys/auth'.format(API_VERSION)).json()

    def enable_auth_backend(self, mount_point, auth_type=None, description=None):
        """
        POST /sys/auth/<mount point>

        https://vaultproject.io/docs/http/sys-auth.html
        """
        if not auth_type:
            auth_type = mount_point

        params = {
            'type': auth_type,
            'description': description,
        }

        self._post('/{}/sys/auth/{}'.format(API_VERSION,
            mount_point), params)

    def disable_auth_backend(self, mount_point):
        """
        DELETE /sys/auth/<mount point>

        https://vaultproject.io/docs/http/sys-auth.html
        """
        self._delete('/{}/sys/auth/{}'.format(API_VERSION,
            mount_point))
